(*
 *  TTAPILine component version 1.0
 *  (c)1998 Jury Kornienko
 *  j.kornienko@pf.lv
 *
 *
 *  See for updates in
 *  http://www.pf.lv
 *
 *  TAPI.pas - Translated by Alexander Staubo
 *
 *  Source code provided under some conditions:
 *
 *  - if you use significant parts of the code, please credit me
 *  - do not release a new version of this component without
 *    significant enhancements
 *
 *)

unit TAPILine;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Tapi, ComCtrls, ExtCtrls, SyncObjs, dsgnintf;

const
  TL_TIMEOUT = 20000;
  HiVer = $00020000;  // Highest API version wanted (2.0)
  LoVer = $00010004;  // Lowest API version accepted (1.4)
  EVENT_TIMEOUT = 300;
  NEW_LINE = Chr (13)+Chr (10);

//Reasons why a line device might not be usable by TapiComm.
  LINENOTUSEABLE_ERROR            =1;
  LINENOTUSEABLE_NOVOICE          =2;
  LINENOTUSEABLE_NODATAMODEM      =3;
  LINENOTUSEABLE_NOMAKECALL       =4;
  LINENOTUSEABLE_ALLOCATED        =5;
  LINENOTUSEABLE_INUSE            =6;
  LINENOTUSEABLE_NOCOMMDATAMODEM  =7;

const
    LINEUNAVAIL = 'Line Unavailable';
    LINEUNNAMED = 'Line Unnamed';
    LINENAMEEMPTY = 'Line Name is Empty';

type

  TAboutTAPILineProperty = class(TPropertyEditor)
  public
  	procedure Edit; override;
  	function GetAttributes: TPropertyAttributes; override;
  	function GetValue: string; override;
  end;

  TformDialing = class(TForm)
    bHangUp: TButton;
    Image1: TImage;
    ModemName: TLabel;
    procedure bHangUpClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

  TTAPILine = class(TComponent)
  private
    { Private declarations }
    FAbout: TAboutTAPILineProperty;
    FDirectAccessLogFile: string;{file name}
    FDirectAccessScript: TStrings;
    FDirectAccess: boolean;
    sDirectAccessResult: string;

    FOnDisconnect: TNotifyEvent;
    FOnBusy: TNotifyEvent;
    FOnConnection: TNotifyEvent;
    FOnCallAnswering: TNotifyEvent;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnDropCall: TNotifyEvent;

    hCommFile: LONG;
    LineApp : HLINEAPP;
    Line : HLINE;
    Call : HCALL;

    FCountryCode: integer;
    FPhoneNumber: string;
    FDevIDToInitialize: DWORD;
    FWaitForCall: boolean;
    FInfoLabel: TStatusPanel;

    anotherData: string;
    Parent: TComponent;
    eventCallInProcess: TEvent;
    FOnCallOffering: TNotifyEvent;{call in}
    dwRequestAnswer: DWORD;
    dwRequestMakeCall: DWORD;
    dwRequestDrop: DWORD;
    Procedure OnConnectionInt;
    procedure SetPhoneNumber (value: String);
    procedure SetCountryCode (value: integer);
    procedure SetDevIDToInitialize (value: DWORD);
    function LineCallStateToStr (CallState : DWORD) : string;
    function LineCallStateInfoToStr (CallState, Info : DWORD) : string;
    procedure SetWaitForCall (value: boolean);
    function ClearCommErrorToStr (dwErrorFlags : DWORD) : string;
    procedure OnReply (dwParam1: DWORD; dwErrorCode: DWORD);
    function I_lineNegotiateLegacyAPIVersion (dwDeviceID: DWORD): DWORD;
    function I_lineNegotiateAPIVersion (dwDeviceID: DWORD): DWORD;
    procedure MakeDirectAccessScript;
    procedure SetLines (Value: TStrings);
    function GetLineName (lpLineDevCaps_: LPLINEDEVCAPS): string;
    procedure WriteLog (tag: string; responce: string; slLog: TStringList);
    procedure SetInfoLabel(Value: TStatusPanel);
  public
    { Public declarations }
    bDropped: boolean;
    Version : DWORD;
    NumDevs : Integer; {Number of TAPI devices}
    bConnected: boolean;
    InfoLabel: TStatusPanel;
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ConfigDialog;
    function Drop: boolean;
    function MakeCall: boolean;
    procedure Open;
    function Write (const sData: String): Boolean;
    function Read(wantedToRead: DWORD):string;
    function ReadLargeData (buffer: PChar; wantedToRead: DWORD):LONG;
    procedure Close;
    procedure TranslateDialog;
    function GetIcon: HICON;
    function FillTAPILine (sl: TStrings): DWORD;
    function VerifyUsableLine (dwDeviceID: DWORD): LONG;
  published
    property About: TAboutTAPILineProperty read FAbout write FAbout;
    property WaitForCall: boolean read FWaitForCall write FWaitForCall;
    property DeviceDirectAccess: boolean read FDirectAccess write FDirectAccess;
    property PhoneNumber: string read FPhoneNumber write SetPhoneNumber;
    property CountryCode: integer read FCountryCode write SetCountryCode;
    property DevIdToInitialize: DWORD read FDevIdToInitialize write SetDevIdToInitialize;
    property DirectAccessScript: TStrings read FDirectAccessScript write SetLines;
    property DirectAccessLogFile: string read FDirectAccessLogFile write FDirectAccessLogFile;
    property OnConnection: TNotifyEvent read FOnConnection write FOnConnection;
    property OnCallOffering: TNotifyEvent read FOnCallOffering write FOnCallOffering;
    property OnCallAnswering: TNotifyEvent read FOnCallAnswering write FOnCallAnswering;
    property OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;
    property OnBusy: TNotifyEvent read FOnBusy write FOnBusy;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnDropCall: TNotifyEvent read FOnDropCall write FOnDropCall;
  end;

   procedure LineCallbackProc (hDevice,
                            dwMessage,
                            dwInstance,
                            dwParam1,
                            dwParam2,
                            dwParam3 : DWORD); stdcall;

   procedure ZeroMem (p: PChar);
   procedure Register;

   {$R TAPILine.dcr}
   Var
      thisTAPILine: TTAPILine;
      formDialing: TformDialing;

implementation

{$R *.DFM}

{
Register the component with the Delphi IDE
}
procedure Register;
begin
    RegisterComponents('Freeware', [TTAPILine]);
    RegisterPropertyEditor(TypeInfo(TAboutTAPILineProperty), TTAPILine, 'ABOUT', TAboutTAPILineProperty);
end;

(*
 * About property
 *)
procedure TAboutTAPILineProperty.Edit;
begin
    Application.MessageBox ('This component is freeware. (c) 1998 Jury Kornienko',
       			    'TTAPILine component version 1.0', MB_OK or MB_ICONINFORMATION);
end;

function TAboutTAPILineProperty.GetAttributes: TPropertyAttributes;
begin
    Result := [paMultiSelect, paDialog, paReadOnly];
end;

function TAboutTAPILineProperty.GetValue: string;
begin
    Result := '(about)';
end;

(*
 * TAPILine
 *)
function TTAPILine.LineCallStateToStr (CallState : DWORD) : string;
begin
  case CallState of
    LINECALLSTATE_IDLE :
      Result:='IDLE';
    LINECALLSTATE_OFFERING :
      Result:='OFFERING';
    LINECALLSTATE_ACCEPTED :
      Result:='ACCEPTED';
    LINECALLSTATE_DIALTONE :
      Result:='DIALTONE';
    LINECALLSTATE_DIALING :
      Result:='DIALING';
    LINECALLSTATE_RINGBACK :
      Result:='RINGBACK';
    LINECALLSTATE_BUSY :
      Result:='BUSY';
    LINECALLSTATE_SPECIALINFO :
      Result:='SPECIALINFO';
    LINECALLSTATE_CONNECTED :
      Result:='CONNECTED';
    LINECALLSTATE_PROCEEDING :
      Result:='PROCEEDING';
    LINECALLSTATE_ONHOLD :
      Result:='ONHOLD';
    LINECALLSTATE_CONFERENCED :
      Result:='CONFERENCED';
    LINECALLSTATE_ONHOLDPENDCONF :
      Result:='ONHOLDPENDCONF';
    LINECALLSTATE_ONHOLDPENDTRANSFER :
      Result:='ONHOLDPENDTRANSFER';
    LINECALLSTATE_DISCONNECTED :
      Result:='DISCONNECTED';
    LINECALLSTATE_UNKNOWN :
      Result:='UNKNOWN';
  else
    Result:='Unknown (' + IntTostr(CallState) + ')';
  end;
end;

function TTAPILine.LineCallStateInfoToStr (CallState, Info : DWORD) : string;
begin
  Result:='';
  if Info = 0 Then
     Result:='SUCCESS';
  case CallState of
    LINECALLSTATE_DIALTONE :
      case Info of
        LINEDIALTONEMODE_NORMAL :
          Result:='NORMAL';
        LINEDIALTONEMODE_SPECIAL :
          Result:='SPECIAL';
        LINEDIALTONEMODE_INTERNAL :
          Result:='INTERNAL';
        LINEDIALTONEMODE_EXTERNAL :
          Result:='EXTERNAL';
        LINEDIALTONEMODE_UNKNOWN :
          Result:='UNKNOWN';
        LINEDIALTONEMODE_UNAVAIL :
          Result:='UNAVAIL';
      end;
    LINECALLSTATE_BUSY :
      case Info of
        LINEBUSYMODE_sTATION :
          Result:='STATION';
        LINEBUSYMODE_TRUNK :
          Result:='TRUNK';
        LINEBUSYMODE_UNKNOWN :
          Result:='UNKNOWN';
        LINEBUSYMODE_UNAVAIL :
          Result:='UNAVAIL';
      end;
    LINECALLSTATE_DISCONNECTED :
      case Info of
        LINEDISCONNECTMODE_NORMAL :
          Result:='NORMAL';
        LINEDISCONNECTMODE_UNKNOWN :
          Result:='UNKNOWN';
        LINEDISCONNECTMODE_REJECT :
          Result:='REJECT';
        LINEDISCONNECTMODE_PICKUP :
          Result:='PICKUP';
        LINEDISCONNECTMODE_FORWARDED :
          Result:='FORWARDED';
        LINEDISCONNECTMODE_BUSY :
          Result:='BUSY';
        LINEDISCONNECTMODE_NOANSWER :
          Result:='NOANSWER';
        LINEDISCONNECTMODE_BADADDRESS :
          Result:='BADADDRESS';
        LINEDISCONNECTMODE_UNREACHABLE :
          Result:='UNREACHABLE';
        LINEDISCONNECTMODE_CONGESTION :
          Result:='CONGESTION';
        LINEDISCONNECTMODE_INCOMPATIBLE :
          Result:='INCOMPATIBLE';
        LINEDISCONNECTMODE_UNAVAIL :
          Result:='UNAVAIL';
        LINEDISCONNECTMODE_NODIALTONE:      // TAPI v1.4
          Result:='NODIALTONE';
      end;
  end;
  if Result = '' then
    Result:='Undefined (' + IntToStr(Info) + ')';
end;

procedure LineCallbackProc (hDevice,
                            dwMessage,
                            dwInstance,
                            dwParam1,
                            dwParam2,
                            dwParam3 : DWORD); stdcall; far;
var
  S : string;
begin
  case dwMessage of
       LINE_CLOSE:
           S := 'Line closed';

       LINE_CALLSTATE:
           Begin
           S := thisTAPILine.LineCallStateInfoToStr (dwParam1, dwParam2);

           case dwParam1 of
                LINECALLSTATE_CONNECTED: // CONNECTED!!!
                  Begin
                  if not thisTAPILine.bDropped Then
                     Begin
                     thisTAPILine.OnConnectionInt;

                     { Perform a direct access to device}
                     if (thisTAPILine.FDirectAccess) and
                         (thisTAPILine.FDirectAccessScript.Count <> 0) then
                         try
                            thisTAPILine.MakeDirectAccessScript;
                         except
                            ShowMessage (thisTAPILine.sDirectAccessResult);
                            formDialing.Hide;
                            exit;
                         End;

                     formDialing.Hide;

                     if Assigned (thisTAPILine.OnConnection) Then thisTAPILine.OnConnection (thisTAPILine);
                     End;
                  End;
                LINECALLSTATE_OFFERING:
                  if thisTAPILine.FWaitForCall and (not thisTAPILine.bConnected) then
                     Begin
                     if Assigned (thisTAPILine.OnCallOffering) Then thisTAPILine.OnCallOffering (thisTAPILine);
                     formDialing.Show;
                     thisTAPILine.Call := HCALL (hDevice);
                     TapiCheck(lineAnswer(thisTAPILine.Call, nil, 0));
                     End;
                LINECALLSTATE_DISCONNECTED:
                  Begin
                  if Assigned (thisTAPILine.OnDisconnect) Then thisTAPILine.OnDisconnect (thisTAPILine);
                  formDialing.Hide;
                  End;
                LINECALLSTATE_BUSY:
                  Begin
                  if Assigned (thisTAPILine.OnBusy) Then thisTAPILine.OnBusy (thisTAPILine);
                  formDialing.Hide;
                  End;
           else
                S := thisTAPILine.LineCallStateToStr (dwParam1);
           end;
           End;
    LINE_REPLY :
           thisTAPILine.OnReply (dwParam1, dwParam2);
  end;

  if Assigned (thisTAPILine.InfoLabel) and
     (S <> '') then 
     thisTAPILine.InfoLabel.Text := S;
end;

procedure TTAPILine.OnReply (dwParam1: DWORD; dwErrorCode: DWORD);
Begin
    if dwParam1 = dwRequestAnswer Then
       Begin
       TapiCheck (dwErrorCode);
       if Assigned (OnCallAnswering) Then OnCallAnswering (self);
       dwRequestAnswer := 0;
       End;

    if dwParam1 = dwRequestMakeCall Then
       Begin
//       if Assigned (dwRequestMakeCall) Then dwRequestMakeCall (self);
       TapiCheck (dwErrorCode);
       dwRequestMakeCall := 0;
       End;

    if dwParam1 = dwRequestDrop Then
       Begin
       TapiCheck (dwErrorCode);
       dwRequestAnswer := 0;
       dwRequestMakeCall := 0;
       formDialing.Hide;
//       if Assigned (dwRequestMakeCall) Then dwRequestMakeCall (self);
       dwRequestDrop := 0;
       End;
End;

function TTAPILine.MakeCall: boolean;
var
  LineCallParams : TLINECALLPARAMS;

begin

  if DevIdToInitialize = MAXDWORD Then
     raise Exception.Create ('TAPILine: You must select TAPI compatible line');

  Result := False;

  if dwRequestMakeCall > 0 Then
     exit;

  formDialing.Show;

  FillChar(LineCallParams, SizeOf(LineCallParams), 0);
  with LineCallParams do
    begin
      dwTotalSize:=SizeOf(LineCallParams);
      if FDirectAccess then
         dwBearerMode:=LINEBEARERMODE_PASSTHROUGH
      else
         dwBearerMode:=LINEBEARERMODE_VOICE;
      dwMediaMode:=LINEMEDIAMODE_DATAMODEM;
    end;

  bDropped := False;
  if FDirectAccess then
     dwRequestMakeCall := lineMakeCall(Line, @Call, nil, 0, @LineCallParams)
  else
     dwRequestMakeCall := lineMakeCall(Line, @Call, PChar(FPhoneNumber), FCountryCode, @LineCallParams);
  TapiCheck (dwRequestMakeCall);

  Result := True;
end;

destructor TTAPILine.Destroy;
begin
  Drop;
  Close;

  TapiCheck(lineShutdown(LineApp));

  eventCallInProcess.Free;

  thisTAPILine := nil;
  inherited Destroy;
end;

constructor TTAPILine.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  if thisTAPILine <> nil then
     raise Exception.Create('TAPILine component already exists in this process.');

  TapiCheck(lineInitialize(@LineApp,
                           GetModuleHandle (nil),
                           LineCallbackProc,
                           PChar (Application.Title),
                           @NumDevs));

  Version:=0;

  if NumDevs = 0 Then
     ShowMessage ('There is no modem installed in your PC.'+#13+#10+
                  'You should install it via Control panel | Modems | Add.');

  FDevIdToInitialize := MAXDWORD;
  FWaitForCall := True;

  InfoLabel := nil;
  Parent := AOwner;
  thisTAPILine := self;
  eventCallInProcess := TEvent.Create (nil, True, True, '');

  hCommFile := INVALID_HANDLE_VALUE;

  if formDialing = nil Then
     formDialing := TformDialing.Create (thisTAPILine);

  FDirectAccess := False;
  FDirectAccessScript := TStringList.Create;
//  FDirectAccessLogFile := '';
End;

procedure TTAPILine.Open;
var
  Ext : TLINEEXTENSIONID;
Begin
  if DevIdToInitialize = MAXDWORD Then
     raise Exception.Create ('TAPILine: You must select TAPI compatible line');

  Version:=0;

  TapiCheck (lineNegotiateAPIVersion(LineApp, DevIdToInitialize,
             LoVer, HiVer, @Version, @Ext));

  TapiCheck (lineOpen (LineApp, DevIdToInitialize, @Line, Version, 0, 0,
             LINECALLPRIVILEGE_OWNER, LINEMEDIAMODE_DATAMODEM, nil));

  if Assigned (thisTAPILine.InfoLabel) then
     thisTAPILine.InfoLabel.Text := 'Line opened';

  thisTAPILine.bDropped := False;
       
  if Assigned (FOnOpen) Then FOnOpen (self);
end;

procedure TTAPILine.ConfigDialog;
begin
  if DevIdToInitialize = MAXDWORD Then
     raise Exception.Create ('TAPILine: You must select TAPI compatible line');

  TapiCheck(lineConfigDialog(DevIdToInitialize, (Parent as TForm).Handle, 'comm/datamodem'));
end;

procedure TTAPILine.TranslateDialog;
Var
   lVer: DWORD;
begin
  if DevIdToInitialize = MAXDWORD Then
     raise Exception.Create ('TAPILine: You must select TAPI compatible line');

  if Version = 0 Then
     lVer := LoVer
  else
     lVer := Version;

    TapiCheck(lineTranslateDialog(LineApp, DevIdToInitialize, lVer, (Parent as TForm).Handle, PChar (PhoneNumber)));
end;

function TTAPILine.Write (const sData: String): Boolean;
var
   dwCharsWritten: DWord;
   overlappedWrite: TOVERLAPPED;
Begin
    Result := False;
    if bConnected Then
       Begin
       Result := True;
       with overlappedWrite do
         Begin
         Internal :=0;
         InternalHigh :=0;
         Offset :=0;
         OffsetHigh :=0;
         hEvent := 0;
         End;

       if LONG(hCommFile) <> INVALID_HANDLE_VALUE then
          begin
          dwCharsWritten := 0;
          WriteFile (LONG(hCommFile), PChar(sData)^, Length(sData), dwCharsWritten, @overlappedWrite);
          if overlappedWrite.offsetHigh = Length(sData) then Result := True;
          End;
       End;
End;

procedure TTAPILine.OnConnectionInt;
Var
   lpVarString_: LPVARSTRING;
   dwSizeofVarString: DWORD;
//   dcbPort: TDCB;
   CommTimeOuts: TCommTimeOuts;
Begin

   // Very first, make sure this isn't a duplicated message.
   // A CALLSTATE message can be sent whenever there is a
   // change to the capabilities of a line, meaning that it is
   // possible to receive multiple CONNECTED messages per call.
   // The CONNECTED CALLSTATE message is the only one in TapiComm
   // where it would cause problems if it where sent more
   // than once.
   if (bConnected) Then
       exit;

   bConnected := TRUE;

   dwSizeofVarString := sizeof(VARSTRING) + 1024;
   hCommFile := 0;

   // Get the handle to the comm port from the driver so we can start
   // communicating.  This is returned in a LPVARSTRING structure.

   // Allocate the VARSTRING structure
   GetMem(LPVOID(lpVarString_), dwSizeofVarString);
   ZeroMem (PChar(lpVarString_));
   lpVarString_^.dwTotalSize:=dwSizeofVarString;

   // Fill the VARSTRING structure
   TapiCheck (lineGetID (Line, 0, Call, LINECALLSELECT_CALL,
              lpVarString_, 'comm/datamodem'));

   // Again, the handle to the comm port is contained in a
   // LPVARSTRING structure.  Thus, the handle is the very first
   // thing after the end of the structure.  Note that the name of
   // the comm port is right after the handle, but I don't want it.
   strmove (@hCommFile, PChar (LONG(lpVarString_) + lpVarString_^.dwStringOffset), sizeof (LONG));

   { Yes, I agree that the name of the comm port is right after
     the handle, but for us will be more suitable to know the length
     of this name, so we have another way.
   This routine is necessary,
         because we want to take modem name
         Let's go! }
{   lpLineDevCaps_ := nil;
   GetMem(LPVOID(lpLineDevCaps_), sizeof(TLINEDEVCAPS) + 1024);
   ZeroMem (PChar(lpLineDevCaps_));
   dwAPIVersion := I_lineNegotiateAPIVersion(DevIdToInitialize);
   lpLineDevCaps_^.dwTotalSize:=sizeof(TLINEDEVCAPS) + 1024;
   TapiCheck (lineGetDevCaps (LineApp, DevIdToInitialize, dwAPIVersion, 0, lpLineDevCaps_));
   formDialing.ModemName.Caption := GetLineName (lpLineDevCaps_);
   FreeMem (lpLineDevCaps_);}

   { We don't need to set the baud rate and other parameters
     because most of them are set up with TAPIserv}
   if (hCommFile <> INVALID_HANDLE_VALUE) then
      begin

         GetCommTimeouts(hCommFile, commtimeouts);

         // The CommTimeout numbers will very likely change if you are
         // coding to meet some kind of specification where
         // you need to reply within a certain amount of time after
         // recieving the last byte.  However,  If 1/4th of a second
         // goes by between recieving two characters, its a good
         // indication that the transmitting end has finished, even
         // assuming a 1200 baud modem.

         commtimeouts.ReadIntervalTimeout         := 250;
         commtimeouts.ReadTotalTimeoutMultiplier  := 0;
         commtimeouts.ReadTotalTimeoutConstant    := 0;
         commtimeouts.WriteTotalTimeoutMultiplier := 0;
         commtimeouts.WriteTotalTimeoutConstant   := 0;

         SetCommTimeouts(hCommFile, commtimeouts);

  //       GetCommMask (hCommFile, fdwEvtMask);
    //     fdwEvtMask := fdwEvtMask or EV_RXCHAR;
      //   SetCommMask (hCommFile, fdwEvtMask);

         PurgeComm (hCommFile, PURGE_TXABORT or PURGE_RXABORT or PURGE_TXCLEAR or PURGE_RXCLEAR);
      end
   else
       bConnected := False;
End;

function TTAPILine.Read(wantedToRead: DWORD):string;
Var
   fReadStat: boolean;
   ComStat: TCOMSTAT;
   dwErrorFlags: DWORD;
   dwError: DWORD;
   overlappedRead: TOVERLAPPED;
   dwReaded, Time: DWORD;
Begin
   if bConnected Then
      Begin
      with overlappedRead do
         Begin
         Internal :=0;
         InternalHigh :=0;
         Offset :=0;
         OffsetHigh :=0;
         hEvent := 0;
         End;

      Time := GetTickCount+TL_TIMEOUT;
      if wantedToRead = 0 Then
         Begin
         // only try to read number of bytes in queue
         dwErrorFlags := 0;
         ClearCommError (hCommFile, dwErrorFlags, @ComStat ) ;
         if dwErrorFlags <> 0 then
            raise Exception.Create('COMM error: ' + ClearCommErrorToStr (dwErrorFlags));

         wantedToRead := ComStat.cbInQue;
         if GetTickCount > Time Then
            raise Exception.Create('TAPILine error: Timeout during read operation');
         End;

      SetLength (Result, wantedToRead);
      SetLength (anotherData, wantedToRead);

      if (wantedToRead > 0) Then
      Begin
          fReadStat := ReadFile (hCommFile, PChar(anotherData)^,
		             wantedToRead, dwReaded, @overlappedRead);
          if (not fReadStat) then
             Begin
             dwError := GetLastError();
             if (dwError = ERROR_IO_PENDING) then
                 Begin
                 // asynchronous i/o is still in progress

                 // We have to wait for read to complete.
                 // This function will timeout according to the
                 // CommTimeOuts.ReadTotalTimeoutConstant variable
                 // Every time it times out, check for port errors
                 while (not GetOverlappedResult (hCommfile,
                       overlappedRead, wantedToRead, TRUE)) do
                     Begin
                     dwError := GetLastError();
                     if (dwError = ERROR_IO_INCOMPLETE) then
                  // normal result if not finished
                         continue
                     else
                       Begin
                       // an error occurred, try to recover
                       ShowMessage ('An error occurred, try to recover');
//                  wsprintf( szError, "<CE-%u>", dwError ) ;
//                  WriteTTYBlock( hWnd, szError, lstrlen( szError ) ) ;
                       ClearCommError( hCommFile, dwErrorFlags, @ComStat ) ;
//                  if ((dwErrorFlags > 0) && DISPLAYERRORS( npTTYInfo ))
                  {
	                  wsprintf( szError, "<CE-%u>", dwErrorFlags ) ;
	                  WriteTTYBlock( hWnd, szError, lstrlen( szError ) ) ;
                  }
                       break;
                       End;
                     End;
                 End
             else
                 Begin
                 // some other error occurred
                 if dwError = ERROR_HANDLE_EOF Then
                 // we're reached the end of the file
	         // during the call to ReadFile
                    ShowMessage ('We''re reached the end of the file')
                 else
                 // code to handle that
                    ShowMessage ('some other error occurred');

                 wantedToRead := 0 ;
                 End;
             End; {if (not fReadStat) then}
         End; {if (wantedToRead > 0) Then}
      End; {if bConnected Then}
   Result := anotherData;
 // end of ReadCommBlock()
End;

procedure TTAPILine.SetPhoneNumber (value: String);
begin
    if FPhoneNumber <> value Then
       FPhoneNumber := value;
end;

procedure TTAPILine.SetCountryCode (value: integer);
begin
    if FCountryCode <> value Then
       FCountryCode := value;
end;

procedure TTAPILine.SetDevIDToInitialize (value: DWORD);
Var
  dwAPIVersion: DWORD;
  lpLineDevCaps_:  LPLINEDEVCAPS;
begin
    if FDevIDToInitialize <> Value Then
       Begin
       if FDevIDToInitialize <> value Then
          if VerifyUsableLine (value) = 0 Then
             FDevIDToInitialize := value;

       {This routine is necessary,
         because we want to take modem name
         Let's go! }
       GetMem(LPVOID(lpLineDevCaps_), sizeof(TLINEDEVCAPS) + 1024);
       ZeroMem (PChar(lpLineDevCaps_));
       dwAPIVersion := I_lineNegotiateAPIVersion(DevIdToInitialize);
       lpLineDevCaps_^.dwTotalSize:=sizeof(TLINEDEVCAPS) + 1024;
       TapiCheck (lineGetDevCaps (LineApp, DevIdToInitialize, dwAPIVersion, 0, lpLineDevCaps_));
       formDialing.ModemName.Caption := GetLineName (lpLineDevCaps_);
       FreeMem (lpLineDevCaps_);
       End;
end;

procedure ZeroMem (p: PChar);
Var
   i: integer;
begin
    for i := 0 to sizeof (p)-1 do p[i]:= Chr (0);
End;

procedure TTAPILine.SetWaitForCall (value: boolean);
begin
    if FWaitForCall <> value Then
       FWaitForCall := value;
end;

function TTAPILine.ReadLargeData (buffer: PChar; wantedToRead: DWORD):LONG;
Var
   fReadStat: boolean;
   ComStat: TCOMSTAT;
   dwErrorFlags: DWORD;
   dwError: DWORD;
   overlappedRead: TOVERLAPPED;
   dwReaded: DWORD;
   Time: DWORD;
Begin
   Result := 0;

   if bConnected Then
      Begin
      with overlappedRead do
         Begin
         Internal :=0;
         InternalHigh :=0;
         Offset :=0;
         OffsetHigh :=0;
         hEvent := 0;
         End;

      // only try to read number of bytes in queue
      dwErrorFlags := 0;
      ClearCommError (hCommFile, dwErrorFlags, @ComStat ) ;
      Time := GetTickCount+TL_TIMEOUT;
      while (ComStat.cbInQue < wantedToRead) do
          begin
          if dwErrorFlags <> 0 then
             raise Exception.Create('COMM error: ' + ClearCommErrorToStr (dwErrorFlags));
          Application.ProcessMessages;
          ClearCommError (hCommFile, dwErrorFlags, @ComStat);
          if GetTickCount > Time Then
             raise Exception.Create('TAPILine error: Timeout during read operation');
          End;

      if (wantedToRead > 0) Then
         Begin
         fReadStat := ReadFile (hCommFile, buffer^,
		             wantedToRead, dwReaded, @overlappedRead);
         if (not fReadStat) then
            Begin
            dwError := GetLastError();
            if (dwError = ERROR_IO_PENDING) then
               Begin
               // asynchronous i/o is still in progress

               // We have to wait for read to complete.
               // This function will timeout according to the
               // CommTimeOuts.ReadTotalTimeoutConstant variable
               // Every time it times out, check for port errors
               while (not GetOverlappedResult (hCommfile,
                     overlappedRead, wantedToRead, TRUE)) do
                     Begin
                     dwError := GetLastError();
                     if (dwError = ERROR_IO_INCOMPLETE) then
                        // normal result if not finished
                        continue
                     else
                        Begin
                        // an error occurred, try to recover
                        ShowMessage ('An error occurred, try to recover');
                        ClearCommError( hCommFile, dwErrorFlags, @ComStat ) ;
                        break;
                        End;
                     End;
               End
            else
               Begin
               // some other error occurred
               if dwError = ERROR_HANDLE_EOF Then
                  // we're reached the end of the file
                  // during the call to ReadFile
                  ShowMessage ('We''re reached the end of the file')
               else
                  // code to handle that
                  ShowMessage ('some other error occurred');

               wantedToRead := 0 ;
               End;
            End;
         End;
      End;
 // end of ReadCommBlock()
End;

function TTAPILine.ClearCommErrorToStr (dwErrorFlags : DWORD) : string;
begin
  Result := '';

  if dwErrorFlags = CE_BREAK Then
      Result:=Result + 'The hardware detected a break condition.';

  if dwErrorFlags = CE_DNS Then
      Result:=Result + 'Windows 95 only: A parallel device is not selected.';

  if dwErrorFlags = CE_FRAME Then
      Result:=Result + 'The hardware detected a framing error.';

  if dwErrorFlags = CE_IOE Then
      Result:=Result + 'An I/O error occurred during communications with the device.';

  if dwErrorFlags = CE_MODE Then
      Result:=Result + 'The requested mode is not supported, or the hFile parameter is invalid. If this value is specified, it is the only valid error.';

  if dwErrorFlags = CE_OOP Then
      Result:=Result + 'Windows 95 only: A parallel device signaled that it is out of paper.';

  if dwErrorFlags = CE_OVERRUN Then
      Result:=Result + 'A character-buffer overrun has occurred. The next character is lost.';

  if dwErrorFlags = CE_PTO Then
      Result:=Result + 'Windows 95 only: A time-out occurred on a parallel device.';

  if dwErrorFlags = CE_RXOVER Then
      Result:=Result + 'An input buffer overflow has occurred. There is either no room in the input buffer, or a character was received after the end-of-file (EOF) character.';

  if dwErrorFlags = CE_RXPARITY Then
      Result:=Result + 'The hardware detected a parity error.';

  if dwErrorFlags = CE_TXFULL Then
      Result:=Result + 'The application tried to transmit a character, but the output buffer was full.';
End;

function TTAPILine.GetIcon: HICON;
Begin
    if DevIdToInitialize = MAXDWORD Then
       raise Exception.Create ('TAPILine: You must select TAPI compatible line');

    TapiCheck (lineGetIcon (DevIdToInitialize, 'comm/datamodem', Result));
End;

function TTAPILine.Drop: boolean;
Var
   Time: DWORD;
begin

  Result := False;
  eventCallInProcess.SetEvent;
  bConnected := False;
  bDropped := True;

  if Call.unused <> 0 then
     begin
     dwRequestDrop := lineDrop (Call, nil, 0);
     TapiCheck (dwRequestDrop);
     if Assigned (thisTAPILine) Then thisTAPILine.InfoLabel.Text := 'Try to drop call...';
     Call.unused:=0;
     end;

  while eventCallInProcess.WaitFor (EVENT_TIMEOUT) = wrTimeout do
         Application.ProcessMessages;

  if Assigned (FOnDropCall) Then FOnDropCall (self);
end;

procedure TTAPILine.Close;
Var
   dw: DWORD;
Begin
   eventCallInProcess.SetEvent;
   bConnected := False;

   if (hCommFile <> INVALID_HANDLE_VALUE) Then
      Begin
      CloseHandle (hCommFile);
      hCommFile := INVALID_HANDLE_VALUE;
      End;

   if NumDevs <> 0 then
       Begin
       dw := lineClose (Line);
       End;

   if Assigned (FOnClose) Then FOnClose (self);
End;

procedure TTAPILine.MakeDirectAccessScript;
Var
   s, s2 : string;
   pc : PChar;
   i, limit: integer;
   waitTime : LONG;
   sResponse: string;
   slLog: TStringList;
Begin
    if DevIdToInitialize = MAXDWORD Then
       raise Exception.Create ('TAPILine: You must select TAPI compatible line');

    slLog := nil;
    if FDirectAccessLogFile <> '' Then
       slLog := TStringList.Create;
    WriteLog ('TAPILine logfile: '+DateTimeToStr (Now), '', slLog);
    WriteLog ('------------------------------------------', '', slLog);

    //SetThreadPriority (GetCurrentThread, THREAD_PRIORITY_HIGHEST);
    limit := FDirectAccessScript.Count;
    Dec (limit);
    sResponse := '';
    for i := 0 to limit do
        Begin
        s := FDirectAccessScript.Strings [i];
        WriteLog ('Write', s+NEW_LINE, slLog);
        Write (s+NEW_LINE);
        while not bDropped do
              Begin
              waitTime := GetTickCount ()+ 1000;
              while (waitTime > GetTickCount ()) do
                    Application.ProcessMessages;
              sResponse := sResponse + Read (0);
              WriteLog ('Read', sResponse, slLog);
              if Pos ('OK', UpperCase (sResponse)) <> 0 Then
                 Begin
                 WriteLog ('Command complete', sResponse, slLog);
                 sResponse := '';
                 if limit <> i Then
                    break; //Go to do the next command from script
                 End
              else
              if Pos ('CONNECT', UpperCase (sResponse)) <> 0 Then
                 Begin
                 // we made a connection!
                 WriteLog ('Connected', sResponse, slLog);
                 if slLog <> nil Then
                    slLog.SaveToFile (FDirectAccessLogFile);
                 slLog.Free;

                 //SetThreadPriority (GetCurrentThread, THREAD_PRIORITY_NORMAL);
                 PurgeComm (hCommFile, PURGE_TXABORT or PURGE_RXABORT or PURGE_TXCLEAR or PURGE_RXCLEAR);
                 exit;
                 End
              else
              if (Pos ('ERROR', UpperCase (sResponse)) <> 0) or
                 (Pos ('NO CARRIAR', UpperCase (sResponse)) <> 0) or
                 (Pos ('NO DIALTONE', UpperCase (sResponse)) <> 0)  or
                 (Pos ('NO ', UpperCase (sResponse)) <> 0) Then
                 Begin
                 WriteLog ('Error', sResponse, slLog);
                 PurgeComm (hCommFile, PURGE_TXABORT or PURGE_RXABORT or PURGE_TXCLEAR or PURGE_RXCLEAR);
                 if slLog <> nil Then
                    slLog.SaveToFile (FDirectAccessLogFile);
                 slLog.Free;
                 //SetThreadPriority (GetCurrentThread, THREAD_PRIORITY_NORMAL);

                 sDirectAccessResult := 'TAPILine: Device direct access script error: "' + s + '".'+NEW_LINE+
                                         sResponse;
                 raise Exception.Create(sDirectAccessResult);
                 End;
              Application.ProcessMessages;
              End;
        End;

    if bDropped Then
       WriteLog ('Dropped', sResponse, slLog);

    if slLog <> nil Then
       Begin
       slLog.SaveToFile (FDirectAccessLogFile);
       slLog.Free;
       End;

    //SetThreadPriority (GetCurrentThread, THREAD_PRIORITY_NORMAL);
End;

procedure TTAPILine.WriteLog (tag: string; responce: string; slLog: TStringList);
Begin
    if slLog <> nil Then
       slLog.Add (tag + ': '+responce);
End;

//
//  FUNCTION: void FillTAPILine(HWND)
//
//  PURPOSE: Fills the 'TAPI Line' control with the available line devices.
//
//  PARAMETERS:
//    hwndDlg - handle to the current "Dial" dialog
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
//    This function enumerates through all the TAPI line devices and
//    queries each for the device name.  The device name is then put into
//    the 'TAPI Line' control.  These device names are kept in order rather
//    than sorted.  This allows "Dial" to know which device ID the user
//    selected just by the knowing the index of the selected string.
//
//    There are default values if there isn't a device name, if there is
//    an error on the device, or if the device name is an empty string.
//    The device name is also checked to make sure it is null terminated.
//
//    Note that a Legacy API Version is negotiated.  Since the fields in
//    the LINEDEVCAPS structure that we are interested in haven't moved, we
//    can negotiate a lower API Version than this sample is designed for
//    and still be able to access the necessary structure members.
//
//    The first line that is usable by TapiComm is selected as the 'default'
//    line.  Also note that if there was a previously selected line, this
//    remains the default line.  This would likely only occur if this
//    function is called after the dialog has initialized once; for example,
//    if a new line is added.
//
//

function TTAPILine.FillTAPILine (sl: TStrings): DWORD;
Var
    dwDeviceID: DWORD;
    dwAPIVersion: DWORD;
    lpLineDevCaps_: LPLINEDEVCAPS;
    sLineName: string;
//    lpszLineName: PChar;
//    long lReturn;
//    dwDefaultDevice: DWORD;
Begin
    Result := MAXDWORD;

    // Make sure the control is empty.  If it isn't,
    // hold onto the currently selected ID and then reset it.
    sl.Clear;

    lpLineDevCaps_ := nil;
    GetMem(LPVOID(lpLineDevCaps_), sizeof(TLINEDEVCAPS) + 1024);
    ZeroMem (PChar(lpLineDevCaps_));
    lpLineDevCaps_^.dwTotalSize:=sizeof(TLINEDEVCAPS) + 1024;

    for dwDeviceID := 0 to NumDevs-1 do
        begin
        dwAPIVersion := I_lineNegotiateLegacyAPIVersion (dwDeviceID);
        if (dwAPIVersion <> 0) Then
           Begin
           TapiCheck (lineGetDevCaps (LineApp, dwDeviceID, dwAPIVersion, 0, lpLineDevCaps_));

           sLineName := GetLineName (lpLineDevCaps_);
           End
        else  // Couldn't NegotiateAPIVersion.  Line is unavail.
            sLineName := LINEUNAVAIL;

        // Put the device name into the control
        sl.Add (sLineName);
        sLineName := '';

        // If this line is usable and we don't have a default initial
        // line yet, make this the initial line.
{        if ((strcomp (lpszLineName, LINEUNAVAIL) <> 0) and
            (Result = MAXDWORD) and
            (VerifyUsableLine (dwDeviceID) = 0)) Then
            Result := dwDeviceID;}

        End; {for dwDeviceID := 0 to dwNumDevs-1 do}

    if (lpLineDevCaps_ <> nil) then
        freemem (lpLineDevCaps_);

End;

function TTAPILine.GetLineName (lpLineDevCaps_: LPLINEDEVCAPS): string;
Var
   lpszLineName: PChar;
Begin
    GetMem(lpszLineName, 200);
    ZeroMem (lpszLineName);
    Result := '';

    if (lpLineDevCaps_ <> nil) Then
       Begin
       if ((lpLineDevCaps_^.dwLineNameSize <> 0) and
           (lpLineDevCaps_^.dwLineNameOffset <> 0) and
           (lpLineDevCaps_^.dwStringFormat = STRINGFORMAT_ASCII)) then
           Begin
           // This is the name of the device.
           getmem (lpszLineName, lpLineDevCaps_^.dwLineNameSize+1);
           strmove (lpszLineName, PChar (LONG(lpLineDevCaps_) + lpLineDevCaps_^.dwLineNameOffset),
                    lpLineDevCaps_^.dwLineNameSize);

           if (lpszLineName[0] <> Chr (0)) Then
               Begin
               // Reverse indented to make this fit

               // Make sure the device name is null terminated.
               if (lpszLineName [lpLineDevCaps_^.dwLineNameSize -1] <> Chr (0)) Then
                  Begin
                  // If the device name is not null terminated, null
                  // terminate it.  Yes, this looses the end character.
                  // Its a bug in the service provider.
                  lpszLineName[lpLineDevCaps_^.dwLineNameSize-1] := Chr (0);
                  End;
                End
           else // Line name started with a NULL.
               StrPCopy (lpszLineName, LINENAMEEMPTY);
           End
       else  // DevCaps doesn't have a valid line name.  Unnamed.
           StrPCopy (lpszLineName, LINEUNNAMED);

       Result := StrPas (lpszLineName);
       freemem (lpszLineName);
       End
    else  // Couldn't GetDevCaps.  Line is unavail.
       Result := LINEUNAVAIL;
End;

//
//  FUNCTION: DWORD I_lineNegotiateLegacyAPIVersion(DWORD)
//
//  PURPOSE: Negotiate an API Version to use for a specific device.
//
//  PARAMETERS:
//    dwDeviceID - device to negotiate an API Version for.
//
//  RETURN VALUE:
//    Returns the API Version to use for this line if successful.
//    Returns 0 if negotiations fall through.
//
//  COMMENTS:
//
//    This wrapper is slightly different from the I_lineNegotiateAPIVersion.
//    This wrapper allows TapiComm to negotiate an API version between
//    1.3 and SAMPLE_TAPI_VERSION.  Normally, this sample is specific to
//    API Version SAMPLE_TAPI_VERSION.  However, there are a few times when
//    TapiComm needs to get information from a service provider, but also knows
//    that a lower API Version would be ok.  This allows TapiComm to recognize
//    legacy service providers even though it can't use them.  1.3 is the
//    lowest API Version a legacy service provider should support.
//
//

function TTAPILine.I_lineNegotiateLegacyAPIVersion (dwDeviceID: DWORD): DWORD;
Var
  LineExtensionID: TLINEEXTENSIONID;
  dwLocalAPIVersion: DWORD;
  dwReturn: DWORD;
Begin

    dwReturn := lineNegotiateAPIVersion (LineApp, dwDeviceID,
             LoVer, HiVer, @dwLocalAPIVersion, @LineExtensionID);

    if (dwReturn = LINEERR_INCOMPATIBLEAPIVERSION) Then
       Begin
       Result := 0;
       exit;
       End;

    TapiCheck (dwReturn);
    Result := dwLocalAPIVersion;
End;

function TTAPILine.I_lineNegotiateAPIVersion (dwDeviceID: DWORD): DWORD;
Var
  LineExtensionID: TLINEEXTENSIONID;
  dwLocalAPIVersion: DWORD;
  dwReturn: DWORD;
Begin

    dwReturn := lineNegotiateAPIVersion (LineApp, dwDeviceID,
             LoVer, LoVer, @dwLocalAPIVersion, @LineExtensionID);

    if (dwReturn = LINEERR_INCOMPATIBLEAPIVERSION) Then
       Begin
       Result := 0;
       exit;
       End;

    TapiCheck (dwReturn);
    Result := dwLocalAPIVersion;
End;

//
//  FUNCTION: long VerifyUsableLine(DWORD)
//
//  PURPOSE: Verifies that a specific line device is useable by TapiComm.
//
//  PARAMETERS:
//    dwDeviceID - The ID of the line device to be verified
//
//  RETURN VALUE:
//    Returns SUCCESS if dwDeviceID is a usable line device.
//    Returns a LINENOTUSEABLE_ constant otherwise.
//
//  COMMENTS:
//
//    VerifyUsableLine takes the give device ID and verifies step by step
//    that the device supports all the features that TapiComm requires.
//
//

function TTAPILine.VerifyUsableLine (dwDeviceID: DWORD): LONG;
Var
    hLine_: HLINE;
    lpLineDevCaps_: LPLINEDEVCAPS;
    lpLineAddressStatus_: LPLINEADDRESSSTATUS;
    lpVarString_: LPVARSTRING;
    dwAPIVersion: DWORD;
    lReturn: LONG;
    hCall_ : HCALL;

Label DeleteBuffers;

Begin
    Result := 0;
    hLine_.unused := 0;
    lpLineDevCaps_ := nil;
    lpLineAddressStatus_ := nil;
    lpVarString_ := nil;

    // The line device must support an API Version that TapiComm does.
    dwAPIVersion := I_lineNegotiateAPIVersion(dwDeviceID);
    if (dwAPIVersion = 0) Then
       begin
       Result := LINENOTUSEABLE_ERROR;

       {h-mm, operator GOTO!!!}
       goto DeleteBuffers;
       End;

    lpLineDevCaps_ := nil;
    GetMem(LPVOID(lpLineDevCaps_), sizeof(TLINEDEVCAPS) + 1024);
    ZeroMem (PChar(lpLineDevCaps_));
    lpLineDevCaps_^.dwTotalSize:=sizeof(TLINEDEVCAPS) + 1024;

    TapiCheck (lineGetDevCaps (LineApp, dwDeviceID, dwAPIVersion, 0, lpLineDevCaps_));

    // Must support LINEBEARERMODE_VOICE
    if ((lpLineDevCaps_^.dwBearerModes and LINEBEARERMODE_VOICE) = 0) then
       Begin
       Result := LINENOTUSEABLE_NOVOICE;
       ShowMessage ('VerifyUsableLine: doesn''t support LINEBEARERMODE_VOICE');
       goto DeleteBuffers;
       End;

    // Must support LINEMEDIAMODE_DATAMODEM
    if ((lpLineDevCaps_^.dwMediaModes and LINEMEDIAMODE_DATAMODEM) = 0) then
       Begin
       Result := LINENOTUSEABLE_NODATAMODEM;
       ShowMessage ('VerifyUsableLine: doesn''t support LINEMEDIAMODE_DATAMODEM');
       goto DeleteBuffers;
       End;

    // Must be able to make calls
    if ((lpLineDevCaps_^.dwLineFeatures and LINEFEATURE_MAKECALL) = 0) Then
       Begin
       Result := LINENOTUSEABLE_NOMAKECALL;
       ShowMessage ('VerifyUsableLine: doesn''t support LINEFEATURE_MAKECALL');
       goto DeleteBuffers;
       End;

    // It is necessary to open the line so we can check if
    // there are any call appearances available.  Other TAPI
    // applications could be using all call appearances.
    // Opening the line also checks for other possible problems.

    lReturn := lineOpen(LineApp, dwDeviceID, @hLine_,
            dwAPIVersion, 0, 0,
            LINECALLPRIVILEGE_NONE, LINEMEDIAMODE_DATAMODEM,
            LPLINECALLPARAMS(0));

    if (lReturn = LINEERR_ALLOCATED) Then
       Begin
       ShowMessage ('VerifyUsableLine: ' +NEW_LINE+
                    'Line is already in use by a non-TAPI app or' +NEW_LINE+
                    'another Service Provider.');
       Result := LINENOTUSEABLE_ALLOCATED;
       goto DeleteBuffers;
       End
    else
       Tapicheck (lReturn);

    lpLineAddressStatus_ := nil;
    GetMem(LPVOID(lpLineAddressStatus_), sizeof(TLINEADDRESSSTATUS) + 1024);
    ZeroMem (PChar(lpLineAddressStatus_));
    lpLineAddressStatus_^.dwTotalSize:=sizeof(TLINEADDRESSSTATUS) + 1024;

    // Get LineAddressStatus to make sure the line isn't already in use.
    TapiCheck (lineGetAddressStatus(hLine_, 0, lpLineAddressStatus_));

    // Are there any available call appearances (ie: is it in use)?
    if ( (lpLineAddressStatus_^.dwAddressFeatures and
               LINEADDRFEATURE_MAKECALL) = 0) Then
       Begin
       ShowMessage ('VerifyUsableLine: ' +NEW_LINE+
                    'Line make call feature is not available.');
       Result := LINENOTUSEABLE_INUSE;
       goto DeleteBuffers;
       End;

    // Make sure the "comm/datamodem" device class is supported
    // Note that we don't want any of the 'extra' information
    // normally returned in the VARSTRING structure.  All we care
    // about is if lineGetID succeeds.

    {GetMem(LPVOID(lpVarString_), sizeof(VARSTRING) + 1024);
    ZeroMem (PChar(lpVarString_));
    lpVarString_^.dwTotalSize:=sizeof(VARSTRING) + 1024;

    // Fill the VARSTRING structure
    TapiCheck (lineGetID (hLine_, 0, hCall_, LINECALLSELECT_LINE,
              lpVarString_, 'comm/datamodem'));
     }
    DeleteBuffers:

    if (hCall_.unused <> 0) then
        lineDrop (hCall_, nil, 0);
    if (hLine_.unused <> 0) then
        lineClose (hLine_);
    if (lpLineAddressStatus_ <> nil) then
        Freemem (lpLineAddressStatus_);
    if (lpLineDevCaps_ <> nil) Then
        Freemem (lpLineDevCaps_);
    if (lpVarString_ <> nil) then
        Freemem(lpVarString_);

End;

procedure TTAPILine.SetLines(Value: TStrings);
begin
    if Value <> nil Then
       FDirectAccessScript.Assign(Value);
end;

procedure TTAPILine.SetInfoLabel(Value: TStatusPanel);
begin
    if Value <> nil Then
       FInfoLabel.Assign(Value);
end;

procedure TformDialing.bHangUpClick(Sender: TObject);
begin
    thisTAPILine.Drop;
    formDialing.Hide;
end;

end.

